from libcpp.pair cimport pair
from libcpp.memory cimport shared_ptr
from libcpp.string cimport string
from libcpp cimport bool
from .cat cimport Cat, CatPair
from .combinator cimport Op, Combinator

cdef extern from "tree.h" namespace "myccg" nogil:
    cdef cppclass Node:
        Cat GetCategory() const
        const int GetLength() const
        shared_ptr[const Node] GetLeftChild() const
        shared_ptr[const Node] GetRightChild() const
        bint IsLeaf() const
        int GetStartOfSpan() const
        string GetWord() const
        int GetHeadId() const
        int GetDependencyLength() const
        bint HeadIsLeft() const
        bint IsUnary() const
        int NumDescendants() const
        int RightNumDescendants() const
        int LeftNumDescendants() const

    ctypedef shared_ptr[const Node] NodeType
    ctypedef pair[NodeType, float] ScoredNode

    cdef cppclass Leaf(Node):
        Leaf(const string&, Cat, int)

    cdef cppclass CTree(Node):
        CTree(Cat, bool, NodeType, NodeType, Op)
        CTree(Cat, NodeType)

    cdef cppclass AUTO:
        AUTO(NodeType tree)
        string Get()

    cdef cppclass Derivation:
        Derivation(NodeType tree, bint feat)
        string Get()

    cdef cppclass JaCCG:
        JaCCG(NodeType tree)
        string Get()

    cdef cppclass PyXML:
        PyXML(NodeType tree, bint feat)
        string Get()

    cdef cppclass Prolog:
        Prolog(NodeType tree)
        string Get()

    cdef cppclass CoNLL:
        CoNLL(NodeType tree)
        string Get()

    string EnResolveCombinatorName(const Node*)
    string JaResolveCombinatorName(const Node*)


cdef class Tree:
    cdef NodeType node_
    cdef public bint suppress_feat
    cdef bytes lang

    @staticmethod
    cdef Tree from_ptr(NodeType node, lang)






cdef class Combinator:
    def __cinit__(self):
        pass

    @staticmethod
    cdef Combinator from_ptr(Op op):
        combinator = Combinator()
        combinator.op_ = op
        return combinator


    def __str__(self):
        return self.op_.ToStr().decode('utf-8')

    def __repr__(self):
        return self.op_.ToStr().decode('utf-8')

    cdef bool _can_apply(self, Category left, Category right):
        return self.op_.CanApply(left.cat_, right.cat_)

    def can_apply(self, left, right):
        if isinstance(left, Category) and isinstance(right, Category):
            return self._can_apply(left, right)
        else:
            return False

    cdef Category _apply(self, Category left, Category right):
        return Category.from_ptr(self.op_.Apply(left.cat_, right.cat_))

    def apply(self, left, right):
        if self.can_apply(left, right):
            return self._apply(left, right)
        else:
            return None

cpdef conjunction():
    return Combinator.from_ptr(<Op>new Conjunction())

cpdef conjunction2():
    return Combinator.from_ptr(<Op>new Conjunction2())

cpdef coordinate():
    return Combinator.from_ptr(<Op>new Coordinate())

cpdef comma_and_verb_phrase_to_adverb():
    return Combinator.from_ptr(<Op>new CommaAndVerbPhraseToAdverb())

cpdef parenthetical_direct_speech():
    return Combinator.from_ptr(<Op>new ParentheticalDirectSpeech())

cpdef remove_punctuation(bool punct_is_left):
    return Combinator.from_ptr(<Op>new RemovePunctuation(punct_is_left))

cpdef remove_punctuation_left():
    return Combinator.from_ptr(<Op>new RemovePunctuationLeft())

cpdef forward_application():
    return Combinator.from_ptr(<Op>new ForwardApplication())

cpdef backward_application():
    return Combinator.from_ptr(<Op>new BackwardApplication())

def slash_check(slashes):
    for slash in slashes:
        if slash not in ['/', '\\']:
            raise RuntimeError(f'slashes must be either of "\\" or "/": [{", ".join(slashes)}]')

cpdef forward_composition(str left, str right, str result):
    slash_check([left, right, result])
    cdef string c_left = left.encode('utf-8')
    cdef string c_right = right.encode('utf-8')
    cdef string c_result = result.encode('utf-8')
    return Combinator.from_ptr(
        <Op>new ForwardComposition(Slash(c_left[0]), Slash(c_right[0]), Slash(c_result[0])))

cpdef backward_composition(str left, str right, str result):
    slash_check([left, right, result])
    cdef string c_left = left.encode('utf-8')
    cdef string c_right = right.encode('utf-8')
    cdef string c_result = result.encode('utf-8')
    return Combinator.from_ptr(
        <Op>new BackwardComposition(Slash(c_left[0]), Slash(c_right[0]), Slash(c_result[0])))

cpdef generalized_forward_composition(str left, str right, str result):
    cdef string c_left = left.encode('utf-8')
    cdef string c_right = right.encode('utf-8')
    cdef string c_result = result.encode('utf-8')
    return Combinator.from_ptr(
        <Op>new GeneralizedForwardComposition(Slash(c_left[0]), Slash(c_right[0]), Slash(c_result[0])))

cpdef generalized_backward_composition(str left, str right, str result):
    cdef string c_left = left.encode('utf-8')
    cdef string c_right = right.encode('utf-8')
    cdef string c_result = result.encode('utf-8')
    return Combinator.from_ptr(
        <Op>new GeneralizedBackwardComposition(Slash(c_left[0]), Slash(c_right[0]), Slash(c_result[0])))

cpdef headfirst_combinator(Combinator combinator):
    return Combinator.from_ptr((<Op>new HeadFirstCombinator(combinator.op_)))

cpdef headfinal_combinator(Combinator combinator):
    return Combinator.from_ptr((<Op>new HeadFinalCombinator(combinator.op_)))

cpdef en_forward_application():
    return Combinator.from_ptr(<Op>new ENForwardApplication())

cpdef en_backward_application():
    return Combinator.from_ptr(<Op>new ENBackwardApplication())

cpdef conjoin():
    return Combinator.from_ptr(<Op>new Conjoin())

cpdef ja_forward_application():
    return Combinator.from_ptr(<Op>new JAForwardApplication())

cpdef ja_backward_application():
    return Combinator.from_ptr(<Op>new JABackwardApplication())

cpdef ja_generalized_backward_composition0(str left, str right, str result, str name):
    cdef string c_left = left.encode('utf-8')
    cdef string c_right = right.encode('utf-8')
    cdef string c_result = result.encode('utf-8')
    cdef string c_name = name.encode('utf-8')
    return Combinator.from_ptr(
        <Op>new JAGeneralizedBackwardComposition0(
            Slash(c_left[0]), Slash(c_right[0]), Slash(c_result[0]), c_name))

cpdef ja_generalized_backward_composition1(str left, str right, str result, str name):
    cdef string c_left = left.encode('utf-8')
    cdef string c_right = right.encode('utf-8')
    cdef string c_result = result.encode('utf-8')
    cdef string c_name = name.encode('utf-8')
    return Combinator.from_ptr(
        <Op>new JAGeneralizedBackwardComposition1(
            Slash(c_left[0]), Slash(c_right[0]), Slash(c_result[0]), c_name))

cpdef ja_generalized_backward_composition2(str left, str right, str result, str name):
    cdef string c_left = left.encode('utf-8')
    cdef string c_right = right.encode('utf-8')
    cdef string c_result = result.encode('utf-8')
    cdef string c_name = name.encode('utf-8')
    return Combinator.from_ptr(
        <Op>new JAGeneralizedBackwardComposition2(
            Slash(c_left[0]), Slash(c_right[0]), Slash(c_result[0]), c_name))

cpdef ja_generalized_backward_composition3(str left, str right, str result, str name):
    cdef string c_left = left.encode('utf-8')
    cdef string c_right = right.encode('utf-8')
    cdef string c_result = result.encode('utf-8')
    cdef string c_name = name.encode('utf-8')
    return Combinator.from_ptr(
        <Op>new JAGeneralizedBackwardComposition3(
            Slash(c_left[0]), Slash(c_right[0]), Slash(c_result[0]), c_name))

cpdef ja_generalized_forward_composition0(str left, str right, str result, str name):
    cdef string c_left = left.encode('utf-8')
    cdef string c_right = right.encode('utf-8')
    cdef string c_result = result.encode('utf-8')
    cdef string c_name = name.encode('utf-8')
    return Combinator.from_ptr(
        <Op>new JAGeneralizedForwardComposition0(
            Slash(c_left[0]), Slash(c_right[0]), Slash(c_result[0]), c_name))

cpdef ja_generalized_forward_composition1(str left, str right, str result, str name):
    cdef string c_left = left.encode('utf-8')
    cdef string c_right = right.encode('utf-8')
    cdef string c_result = result.encode('utf-8')
    cdef string c_name = name.encode('utf-8')
    return Combinator.from_ptr(
        <Op>new JAGeneralizedForwardComposition1(
            Slash(c_left[0]), Slash(c_right[0]), Slash(c_result[0]), c_name))

cpdef ja_generalized_forward_composition2(str left, str right, str result, str name):
    cdef string c_left = left.encode('utf-8')
    cdef string c_right = right.encode('utf-8')
    cdef string c_result = result.encode('utf-8')
    cdef string c_name = name.encode('utf-8')
    return Combinator.from_ptr(
        <Op>new JAGeneralizedForwardComposition2(
            Slash(c_left[0]), Slash(c_right[0]), Slash(c_result[0]), c_name))

cpdef remove_disfluency():
    return Combinator.from_ptr(<Op>new RemoveDisfluency())

cpdef unknown_combinator():
    return Combinator.from_ptr(<Op>new UnknownCombinator())


cdef vector[Op] combinator_list_to_vector(list combinators):
    cdef vector[Op] results
    cdef Combinator combinator
    for combinator in combinators:
        results.push_back(combinator.op_)
    return results


en_default_binary_rules = [
    headfirst_combinator(en_forward_application()),
    headfirst_combinator(en_backward_application()),
    headfirst_combinator(forward_composition('/', '/', '/')),
    headfirst_combinator(backward_composition('/', '\\', '/')),
    headfirst_combinator(generalized_forward_composition('/', '/', '/')),
    headfirst_combinator(generalized_backward_composition('/', '/', '/')),
    headfirst_combinator(conjunction()),
    headfirst_combinator(conjunction2()),
    headfirst_combinator(remove_punctuation(False)),
    headfirst_combinator(remove_punctuation(True)),
    headfirst_combinator(remove_punctuation_left()),
    headfirst_combinator(comma_and_verb_phrase_to_adverb()),
    headfirst_combinator(parenthetical_direct_speech())
]


ja_default_binary_rules = [
    headfinal_combinator(conjoin()),
    headfinal_combinator(ja_forward_application()),
    headfinal_combinator(ja_backward_application()),
    headfinal_combinator(ja_generalized_forward_composition0('/', '/', '/', '>B')),
    headfinal_combinator(ja_generalized_backward_composition0('\\', '\\', '\\', '<B1')),
    headfinal_combinator(ja_generalized_backward_composition1('\\', '\\', '\\', '<B2')),
    headfinal_combinator(ja_generalized_backward_composition2('\\', '\\', '\\', '<B3')),
    headfinal_combinator(ja_generalized_backward_composition3('\\', '\\', '\\', '<B4')),
    headfinal_combinator(ja_generalized_forward_composition0('/', '\\', '\\', '>Bx1')),
    headfinal_combinator(ja_generalized_forward_composition1('/', '\\', '\\', '>Bx2')),
    headfinal_combinator(ja_generalized_forward_composition2('/', '\\', '\\', '>Bx3')),
]



from libcpp.vector cimport vector
from libcpp.string cimport string
from libcpp.unordered_set cimport unordered_set
from libcpp.unordered_map cimport unordered_map
from libcpp cimport bool
from .cat cimport Cat, CatPair


cdef vector[Cat] cat_list_to_vector(list cats)

cdef unordered_set[Cat] cat_list_to_unordered_set(list cats)

cdef unordered_map[string, unordered_set[Cat]] convert_cat_dict(dict cat_dict)

cdef unordered_map[Cat, unordered_set[Cat]] convert_unary_rules(list unary_rules)

cpdef read_unary_rules(filename)

cpdef read_cat_dict(filename)

cpdef read_cat_list(filename)

cpdef read_seen_rules(filename, preprocess)

cdef unordered_set[CatPair] convert_seen_rules(seen_rule_list)

cdef unordered_set[Cat] read_possible_root_categories(list cats)




from libcpp.vector cimport vector
from libcpp.string cimport string
from libcpp.unordered_set cimport unordered_set
from libcpp.unordered_map cimport unordered_map
from libcpp cimport bool
# from libcpp.functional cimport function
from .cat cimport Category, Cat, CatPair
from .tree cimport Tree, ScoredNode, NodeType
from .combinator cimport Op


cdef extern from "depccg.h" namespace "myccg" nogil:
    cdef cppclass RuleCache

    cdef cppclass AgendaItem

    cdef cppclass PartialConstraints:
        PartialConstraints()
        PartialConstraints(const unordered_map[Cat, unordered_set[Cat]]& unary_rules)
        void Add(Cat, unsigned, unsigned)
        void Add(unsigned, unsigned)

    ctypedef vector[NodeType] (*ApplyBinaryRules)(
            const unordered_set[CatPair]&, NodeType, NodeType, unsigned, unsigned)

    ctypedef unordered_set[Cat] (*ApplyUnaryRules)(
            const unordered_map[Cat, unordered_set[Cat]]&, NodeType)

    ApplyUnaryRules DefaultApplyUnaryRules

    ApplyBinaryRules MakeDefaultApplyBinaryRules(const vector[Op]&)

    ApplyUnaryRules EnApplyUnaryRules

    ApplyBinaryRules MakeEnApplyBinaryRules(const vector[Op]&)

    vector[ScoredNode] ParseSentence(
            unsigned id,
            const string& sent,
            float* tag_scores,
            float* dep_scores,
            const unordered_map[string, unordered_set[Cat]]& category_dict,
            const vector[Cat]& tag_list,
            float unary_penalty,
            float beta,
            bool use_beta,
            unsigned pruning_size,
            unsigned nbest,
            const unordered_set[Cat]& possible_root_cats,
            const unordered_map[Cat, unordered_set[Cat]]& unary_rules,
            const unordered_set[CatPair]& seen_rules,
            ApplyBinaryRules apply_binary_rules,
            ApplyUnaryRules apply_unary_rules,
            PartialConstraints constraints,
            unsigned max_length,
            unsigned max_steps) nogil

    vector[vector[ScoredNode]] ParseSentences(
            vector[string]& sents,
            float** tag_scores,
            float** dep_scores,
            const unordered_map[string, unordered_set[Cat]]& category_dict,
            const vector[Cat]& tag_list,
            float unary_penalty,
            float beta,
            bint use_beta,
            unsigned pruning_size,
            unsigned nbest,
            const unordered_set[Cat]& possible_root_cats,
            const unordered_map[Cat, unordered_set[Cat]]& unary_rules,
            const unordered_set[CatPair]& seen_rules,
            const vector[ApplyBinaryRules]& apply_binary_rules,
            ApplyUnaryRules apply_unary_rules,
            vector[PartialConstraints]& constraints,
            unsigned max_length,
            unsigned max_steps)


import os
import subprocess
import tempfile
from pathlib import Path
from typing import List
from lxml import etree

from depccg.download import MODEL_DIRECTORY
from depccg.morpha import MorphaStemmer
from depccg.printer import logger


class Token(dict):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def __getattr__(self, item):
        return self[item]

    def __repr__(self):
        res = super().__repr__()
        return f'Token({res})'

    @classmethod
    def from_piped(cls, string: str) -> 'Token':
        # WORD|POS|NER or WORD|LEMMA|POS|NER
        # or WORD|LEMMA|POS|NER|CHUCK
        items = string.split('|')
        if len(items) == 5:
            word, lemma, pos, entity, chunk = items
        elif len(items) == 4:
            word, lemma, pos, entity = items
            chunk = 'XX'
        else:
            assert len(items) == 3
            word, pos, entity = items
            lemma = 'XX'
            chunk = 'XX'

        return Token(word=word,
                     lemma=lemma,
                     pos=pos,
                     entity=entity,
                     chunk=chunk)

    @classmethod
    def from_word(cls, word: str) -> 'Token':
        return Token(word=word,
                     lemma='XX',
                     pos='XX',
                     entity='XX',
                     chunk='XX')


candc_cmd = "cat \"{0}\" | {1}/bin/pos --model {2} | {1}/bin/ner --model {3}"


def annotate_XX(sentences: List[List[str]], tokenize=False) -> List[List[Token]]:
    if tokenize:
        raise NotImplementedError('no tokenizer implemented')

    return [[Token.from_word(word) for word in sentence]
            for sentence in sentences]


def try_annotate_using_candc(sentences: List[List[str]], tokenize=False) -> List[List[Token]]:
    if tokenize:
        raise NotImplementedError('no tokenizer implemented in the C&C pipeline')

    candc_dir = os.environ.get('CANDC', None)
    candc_model_pos = None
    candc_model_ner = None
    fail = False
    if candc_dir:
        candc_dir = Path(candc_dir)
        candc_model_pos = Path(os.environ.get('CANDC_MODEL_POS', str(candc_dir / 'models' / 'pos')))
        candc_model_ner = Path(os.environ.get('CANDC_MODEL_NER', str(candc_dir / 'models' / 'ner')))
        if (candc_dir / 'bin' / 'pos').exists() and \
                (candc_dir / 'bin' / 'ner').exists() and \
                candc_model_pos.exists() and \
                candc_model_ner.exists():
            pass
        else:
            logger.info('CANDC environmental variable may not be configured correctly.')
            logger.info('$CANDC/bin/{pos,ner} and $CANDC/models/{pos,ner} are expected to exist.')
            fail = True
    else:
        fail = True

    if fail:
        logger.info('did not find C&C parser at CANDC environmental variable.')
        logger.info('fill POS tag etc. using XX tag.')
        return annotate_XX(sentences)

    logger.info('find C&C parser at CANDC environmental variable.')
    logger.info('use C&C pipeline to annotate POS and NER infos.')
    logger.info(f'C&C models: [{candc_model_pos}, {candc_model_ner}]')

    stemmer = MorphaStemmer(str(MODEL_DIRECTORY / 'verbstem.list'))

    tmpfile = tempfile.mktemp()
    with open(tmpfile, 'w') as f:
        for sentence in sentences:
            print(' '.join(sentence), file=f)

    command = candc_cmd.format(tmpfile,
                               candc_dir,
                               candc_model_pos,
                               candc_model_ner)
    proc = subprocess.Popen(command,
                            shell=True,
                            stdin=subprocess.PIPE,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE)

    res, error = proc.communicate()
    try:
        tagged_sentences = res.decode('utf-8').strip().split('\n')
        tagged_sentences = [[tuple(token.split('|')) for token in sentence.strip().split(' ')]
                            for sentence in tagged_sentences]
    except:
        raise RuntimeError('failed to process C&C output. there might have been some problem '
                           'during running C&C pipeline?\n'
                           f'stderr:\n {error}')

    res = []
    for sentence in tagged_sentences:
        words, poss = zip(*[(word, pos) for word, pos, _ in sentence])
        lemmas = stemmer.analyze(list(words), list(poss))
        tokens = [Token(word=word, pos=pos, entity=ner, lemma=lemma.lower(), chunk='XX')
                  for (word, pos, ner), lemma in zip(sentence, lemmas)]
        res.append(tokens)
    return res


def annotate_using_spacy(sentences, tokenize=False, n_threads=2, batch_size=10000):
    try:
        import spacy
    except ImportError:
        logger.error('failed to import janome. please install it by "pip install janome".')
        exit(1)

    nlp = spacy.load('en', disable=['parser'])
    logger.info('use spacy to annotate POS and NER infos.')

    if tokenize:
        docs = [nlp.tokenizer(' '.join(sentence)) for sentence in sentences]
        raw_sentences = [[str(token) for token in doc] for doc in docs]
    else:
        docs = [nlp.tokenizer.tokens_from_list(sentence) for sentence in sentences]
    for name, proc in nlp.pipeline:
        docs = proc.pipe(docs,
                         n_threads=n_threads,
                         batch_size=batch_size)

    res = []
    for sentence in docs:
        tokens = []
        for token in sentence:
            if token.ent_iob_ == 'O':
                ner = token.ent_iob_
            else:
                ner = token.ent_iob_ + '-' + token.ent_type_

            # takes care of pronoun
            if token.lemma_ == '-PRON-':
                lemma = str(token).lower()
            else:
                lemma = token.lemma_.lower()
            tokens.append(
                Token(word=str(token),
                      pos=token.tag_,
                      entity=ner,
                      lemma=lemma,
                      chunk='XX'))
        res.append(tokens)
    if tokenize:
        return res, raw_sentences
    else:
        return res


def annotate_using_janome(sentences, tokenize=False):
    assert tokenize, 'no support for using janome with pre-tokenized inputs'
    try:
        from janome.tokenizer import Tokenizer
    except ImportError:
        logger.error('failed to import janome. please install it by "pip install janome".')
        exit(1)

    logger.info('use Janome to tokenize and annotate POS infos.')
    tokenizer = Tokenizer()
    res = []
    raw_sentences = []
    for sentence in sentences:
        sentence = ''.join(sentence)
        tokenized = tokenizer.tokenize(sentence)
        tokens = []
        for token in tokenized:
            pos, pos1, pos2, pos3 = token.part_of_speech.split(',')
            token = Token(surf=token.surface,
                          pos=pos,
                          pos1=pos1,
                          pos2=pos2,
                          pos3=pos3,
                          inflectionForm=token.infl_form,
                          inflectionType=token.infl_type,
                          reading=token.reading,
                          base=token.base_form)
            tokens.append(token)
        raw_sentence = [token.surface for token in tokenized]
        res.append(tokens)
        raw_sentences.append(raw_sentence)
    return res, raw_sentences


jigg_cmd = "java -Xmx2g -cp \"{0}/jar/*\" jigg.pipeline.Pipeline -annotators {1} -file {2} -output {3}"


def annotate_using_jigg(sentences, tokenize=False, pipeline='ssplit,kuromoji'):
    assert tokenize, 'no support for using jigg with pre-tokenized inputs'
    logger.info('use Jigg to tokenize and annotate POS infos.')

    jigg_dir = os.environ.get('JIGG', None)
    if not jigg_dir:
        logger.error('did not find Jigg at JIGG environmental variable. exiting..')
        exit(1)

    tmpfile = tempfile.mktemp()
    with open(tmpfile, 'w') as f:
        for sentence in sentences:
            print(' '.join(sentence), file=f)

    outfile = tempfile.mktemp()
    command = jigg_cmd.format(jigg_dir,
                              pipeline,
                              tmpfile,
                              outfile)
    proc = subprocess.Popen(command,
                            shell=True,
                            stdin=subprocess.PIPE,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE)

    proc.communicate()
    res = []
    raw_sentences = []
    for sentence in etree.parse(outfile).getroot().xpath('*//sentence'):
        tokens = []
        for token in sentence.xpath('*//token'):
            attrib = token.attrib
            token = Token(surf=attrib['surf'],
                          pos=attrib['pos'],
                          pos1=attrib['pos1'],
                          pos2=attrib['pos2'],
                          pos3=attrib['pos3'],
                          inflectionForm=attrib['inflectionForm'],
                          inflectionType=attrib['inflectionType'],
                          reading=attrib['reading'],
                          base=attrib['base'])
            tokens.append(token)
        res.append(tokens)
        raw_sentence = [token.surf for token in tokens]
        raw_sentences.append(raw_sentence)
    return res, raw_sentences


english_annotator = {
    'candc': try_annotate_using_candc,
    'spacy': annotate_using_spacy,
}


japanese_annotator = {
    'janome': annotate_using_janome,
    'jigg': annotate_using_jigg
}




from ..tree import Tree
from ..cat import Category
from ..combinator import unknown_combinator
from ..token import Token
from lxml import etree
import logging

logger = logging.getLogger(__name__)


def read_auto(filename, lang='en'):
    for line in open(filename):
        line = line.strip()
        if len(line) == 0:
            continue
        if line.startswith("ID"):
            name = line
        else:
            tree = Tree.of_auto(line, lang)
            yield name, Token(), tree


def read_xml(filename, lang='en'):
    unknown_op = unknown_combinator()  # TODO

    def parse(tree):
        def rec(node):
            attrib = node.attrib
            if node.tag == 'rule':
                cat = Category.parse(attrib['cat'])
                children = [rec(child) for child in node.getchildren()]
                if len(children) == 1:
                    return Tree.make_unary(cat, children[0], lang)
                else:
                    assert len(children) == 2
                    left, right = children
                    return Tree.make_binary(
                        cat, True, left, right, unknown_op, lang)
            else:
                assert node.tag == 'lf'
                cat = Category.parse(attrib['cat'])
                word = attrib['word']
                position = len(tokens)
                token = Token(word=attrib['word'],
                              pos=attrib['pos'],
                              entity=attrib['entity'],
                              lemma=attrib['lemma'],
                              chunk=attrib['chunk'])
                tokens.append(token)
                return Tree.make_terminal(word, cat, position, lang)
        tokens = []
        tree = rec(tree)
        return tree, tokens

    trees = etree.parse(filename).getroot().xpath('ccg')
    for tree in trees:
        name = '_'.join(f'{k}={v}' for k, v in tree.items())
        yield (name,) + parse(tree[0])


def read_jigg_xml(filename, lang='en'):
    unknown_op = unknown_combinator()  # TODO

    def parse(tree, tokens):
        def rec(node):
            attrib = node.attrib
            if 'terminal' not in attrib:
                cat = Category.parse(attrib['category'])
                children = [rec(spans[child]) for child in attrib['child'].split(' ')]
                if len(children) == 1:
                    return Tree.make_unary(cat, children[0], lang)
                else:
                    assert len(children) == 2
                    left, right = children
                    return Tree.make_binary(
                        cat, True, left, right, unknown_op, lang)
            else:
                cat = Category.parse(attrib['category'])
                word = tokens[attrib['terminal']].word
                position = int(attrib['begin'])
                return Tree.make_terminal(word, cat, position, lang)

        spans = {span.attrib['id']: span for span in tree.xpath('./span')}
        return rec(spans[tree.attrib['root']])

    trees = etree.parse(filename).getroot()
    sentences = trees[0][0].xpath('sentence')
    for sentence in sentences:
        token_and_ids = []
        for token in sentence.xpath('.//token'):
            token_attribs = dict(token.attrib)
            token_id = token_attribs['id']
            for no_need in ['id', 'start', 'cat']:
                if no_need in token_attribs:
                    del token_attribs[no_need]
            token_and_ids.append((token_id, Token(**token_attribs)))
        tokens = [token for _, token in token_and_ids]
        for ccg in sentence.xpath('./ccg'):
            tree = parse(ccg, dict(token_and_ids))
            yield ccg.attrib['id'], tokens, tree


def read_trees_guess_extension(filename, lang='en'):
    logger.info(f'reading trees from: {filename}')
    if filename.endswith('.jigg.xml'):
        logger.info(f'read it as jigg XML file')
        yield from read_jigg_xml(filename, lang=lang)
    elif filename.endswith('.xml'):
        logger.info(f'read it as C&C XML file')
        yield from read_xml(filename, lang=lang)
    else:
        logger.info(f'read it as AUTO file')
        yield from read_auto(filename, lang=lang)



from libcpp.string cimport string
from libcpp.pair cimport pair
from libcpp.unordered_map cimport unordered_map
from libcpp cimport bool

cdef extern from "feat.h" namespace "myccg" nogil:
    ctypedef const Feature* Feat
    cdef cppclass Feature:
        @staticmethod
        Slash Fwd()

        @staticmethod
        Slash Bwd()

        string ToStr() const
        bint IsEmpty() const
        bint Matches(Feat other) const
        bint ContainsWildcard() const
        string SubstituteWildcard(const string& string) const
        bint ContainsKeyValue(const string& key, const string& value) const
        Feat ToMultiValue() const
        unordered_map[string, string] Values() const;


cdef extern from "cat.h" namespace "myccg" nogil:
    cdef cppclass Slash:
        Slash(char slash)
        bint IsForward() const
        bint IsBackward() const
        string ToStr() const

    ctypedef const CCategory* Cat
    ctypedef pair[Cat, Cat] CatPair
    cdef cppclass CCategory:
        bool operator==(const CCategory&)
        int GetId() const
        @staticmethod
        Cat Parse(const string& cat)
        string ToStr()
        string ToStrWithoutFeat()

        Cat StripFeat() const
        Cat StripFeat(string& f1) const
        const string& GetType() const
        Feat GetFeat() const
        Cat GetLeft() const
        Cat GetRight() const

        Slash GetSlash() const

        const string WithBrackets() const
        bint IsModifier() const
        bint IsModifierWithoutFeat() const
        bint IsTypeRaised() const
        bint IsTypeRaisedWithoutFeat() const
        bint IsForwardTypeRaised() const
        bint IsBackwardTypeRaised() const
        bint IsFunctor() const
        bint IsPunct() const
        bint IsNorNP() const
        int NArgs() const
        bint Matches(Cat other) const
        Cat Arg(int argn) except +
        Cat LeftMostArg() const
        bint IsFunctionInto(Cat cat) const
        Cat ToMultiValue() const



cdef class Category:
    cdef Cat cat_

    @staticmethod
    cdef Category from_ptr(Cat cat)

    cdef bool equals_to(self, Category other)

    cdef bint _matches(self, Category other)

    cdef bint _is_function_into(self, Category cat)

    cpdef strip_feat(self, feat)




from libcpp.string cimport string
from libcpp.unordered_set cimport unordered_set
from libcpp.vector cimport vector
from .cat cimport Cat, Slash, Category
from libcpp cimport bool

cdef extern from "combinator.h" namespace "myccg" nogil:
    cdef enum RuleType:
        FA      = 0,
        BA      = 1,
        FC      = 2,
        BC      = 3,
        GFC     = 4,
        GBC     = 5,
        FX      = 6,
        BX      = 7,
        CONJ    = 8,
        CONJ2   = 9,
        RP      = 10,
        LP      = 11,
        NOISE   = 12,
        UNARY   = 13,
        LEXICON = 14,
        NONE    = 15,
        SSEQ    = 16,
        F_MOD   = 17,
        B_MOD   = 18,
        FWD_TYPERAISE = 19,
        BWD_TYPERAISE = 20,
        COORD = 21

    cdef cppclass CCombinator:
        string ToStr() const
        bool CanApply(Cat left, Cat right)
        Cat Apply(Cat left, Cat right) const

    ctypedef const CCombinator* Op

    cdef cppclass Conjunction2:
        Conjunction2()

    cdef cppclass Coordinate:
        Coordinate()

    cdef cppclass Conjunction:
        Conjunction()

    cdef cppclass CommaAndVerbPhraseToAdverb:
        CommaAndVerbPhraseToAdverb()

    cdef cppclass ParentheticalDirectSpeech:
        ParentheticalDirectSpeech()

    cdef cppclass RemovePunctuation:
        RemovePunctuation(bool punct_is_left)

    cdef cppclass RemovePunctuationLeft:
        RemovePunctuationLeft()

    cdef cppclass ForwardApplication:
        ForwardApplication()

    cdef cppclass BackwardApplication:
        BackwardApplication()

    cdef cppclass ForwardComposition "myccg::GeneralizedForwardComposition<0, myccg::FC>":
        ForwardComposition(const Slash& left, const Slash& right, const Slash& result)

    cdef cppclass GeneralizedForwardComposition "myccg::GeneralizedForwardComposition<1, myccg::GFC>":
        GeneralizedForwardComposition(const Slash& left, const Slash& right, const Slash& result)

    cdef cppclass BackwardComposition "myccg::GeneralizedBackwardComposition<0, myccg::BC>":
        BackwardComposition(const Slash& left, const Slash& right, const Slash& result)

    cdef cppclass GeneralizedBackwardComposition "myccg::GeneralizedBackwardComposition<1, myccg::GBC>":
        GeneralizedBackwardComposition(const Slash& left, const Slash& right, const Slash& result)

    cdef cppclass HeadFirstCombinator:
        HeadFirstCombinator(Op)

    cdef cppclass HeadFinalCombinator:
        HeadFinalCombinator(Op)

    cdef cppclass ENBackwardApplication:
        ENBackwardApplication()

    cdef cppclass ENForwardApplication:
        ENForwardApplication()

    cdef cppclass Conjoin:
        Conjoin()

    cdef cppclass JAForwardApplication:
        JAForwardApplication()

    cdef cppclass JABackwardApplication:
        JABackwardApplication()


    cdef cppclass JAGeneralizedForwardComposition0 "myccg::JAGeneralizedForwardComposition<0, myccg::FX>":
        JAGeneralizedForwardComposition0(const Slash&, const Slash&, const Slash&, const string& string)

    cdef cppclass JAGeneralizedForwardComposition1 "myccg::JAGeneralizedForwardComposition<1, myccg::FX>":
        JAGeneralizedForwardComposition1(const Slash&, const Slash&, const Slash&, const string& string)

    cdef cppclass JAGeneralizedForwardComposition2 "myccg::JAGeneralizedForwardComposition<2, myccg::FX>":
        JAGeneralizedForwardComposition2(const Slash&, const Slash&, const Slash&, const string& string)


    cdef cppclass JAGeneralizedBackwardComposition0 "myccg::JAGeneralizedBackwardComposition<0, myccg::BC>":
        JAGeneralizedBackwardComposition0(const Slash&, const Slash&, const Slash&, const string& string)

    cdef cppclass JAGeneralizedBackwardComposition1 "myccg::JAGeneralizedBackwardComposition<1, myccg::BC>":
        JAGeneralizedBackwardComposition1(const Slash&, const Slash&, const Slash&, const string& string)

    cdef cppclass JAGeneralizedBackwardComposition2 "myccg::JAGeneralizedBackwardComposition<2, myccg::BC>":
        JAGeneralizedBackwardComposition2(const Slash&, const Slash&, const Slash&, const string& string)

    cdef cppclass JAGeneralizedBackwardComposition3 "myccg::JAGeneralizedBackwardComposition<3, myccg::BC>":
        JAGeneralizedBackwardComposition3(const Slash&, const Slash&, const Slash&, const string& string)


    cdef cppclass RemoveDisfluency:
        RemoveDisfluency()

    cdef cppclass UnknownCombinator:
        UnknownCombinator()


cdef class Combinator:
    cdef Op op_

    @staticmethod
    cdef Combinator from_ptr(Op op)

    cdef bool _can_apply(self, Category left, Category right)

    cdef Category _apply(self, Category left, Category right)


cdef vector[Op] combinator_list_to_vector(list combinators)




import argparse
import sys
import logging
import json
from lxml import etree

from .parser import EnglishCCGParser, JapaneseCCGParser
from .printer import to_mathml, to_prolog, to_xml, to_jigg_xml
from depccg.token import Token, english_annotator, japanese_annotator, annotate_XX
from .download import download, load_model_directory, SEMANTIC_TEMPLATES, CONFIGS
from .utils import read_partial_tree, read_weights
from .combinator import en_default_binary_rules, ja_default_binary_rules
from .combinator import remove_disfluency, headfirst_combinator
from .semantics.ccg2lambda import parse as ccg2lambda
from .semantics.ccg2lambda import nltk2json

Parsers = {'en': EnglishCCGParser, 'ja': JapaneseCCGParser}


def main(args):
    logging.basicConfig(format='%(asctime)s - %(levelname)s - %(name)s - %(message)s',
                        level=logging.CRITICAL if args.silent else logging.INFO)

    if args.weights is not None:
        probs, tag_list = read_weights(args.weights)
    else:
        probs, tag_list = None, None

    if args.lang == 'en':
        binary_rules = en_default_binary_rules
        if args.format in ['ccg2lambda', 'jigg_xml_ccg2lambda']:
            assert args.annotator, \
                f'Specify --annotator argument in using "{args.format}" output format'
        annotate_fun = english_annotator.get(args.annotator, annotate_XX)

    elif args.lang == 'ja':
        binary_rules = ja_default_binary_rules
        if args.format in ['ccg2lambda', 'jigg_xml_ccg2lambda']:
            assert args.tokenize, \
                f'Specify --tokenize argument in using "{args.format}" output format'
        if args.tokenize:
            annotate_fun = japanese_annotator[args.annotator]
        else:
            annotate_fun = annotate_XX
    else:
        assert False

    if args.disfluency:
        assert args.lang == 'en', f'not supported disfluency detection in language: {args.lang}'
        binary_rules.append(headfirst_combinator(remove_disfluency()))
         
    if args.root_cats is not None:
        args.root_cats = args.root_cats.split(',')

    kwargs = dict(
        unary_penalty=args.unary_penalty,
        nbest=args.nbest,
        binary_rules=binary_rules,
        possible_root_cats=args.root_cats,
        pruning_size=args.pruning_size,
        beta=args.beta,
        use_beta=not args.disable_beta,
        use_seen_rules=not args.disable_seen_rules,
        use_category_dict=not args.disable_category_dictionary,
        max_length=args.max_length,
        max_steps=args.max_steps,
        gpu=args.gpu
    )

    use_allennlp = args.model and args.model.endswith('.tar.gz')
    if use_allennlp:
        config = CONFIGS[args.lang]
        parser = Parsers[args.lang].from_json(config, args.model, **kwargs)
    else:
        load_tagger = True  # args.input_format != 'json'
        model = args.model or load_model_directory(args.lang)
        parser = Parsers[args.lang].from_dir(model, load_tagger=load_tagger, **kwargs)

    fin = sys.stdin if args.input is None else open(args.input)

    if args.input_format == 'POSandNERtagged':
        tagged_doc = [[Token.from_piped(token) for token in sent.strip().split(' ')] for sent in fin]
        doc = [' '.join(token.word for token in sent) for sent in tagged_doc]
        res = parser.parse_doc(doc,
                               probs=probs,
                               tag_list=tag_list,
                               batchsize=args.batchsize)
    elif args.input_format == 'json':
        doc = [json.loads(line) for line in fin]
        tagged_doc = annotate_fun(
            [[word for word in sent['words'].split(' ')] for sent in doc])
        res = parser.parse_json(doc)
    elif args.input_format == 'partial':
        doc, constraints = zip(*[read_partial_tree(l.strip()) for l in fin])
        tagged_doc = annotate_fun(doc)
        res = parser.parse_doc(doc,
                               probs=probs,
                               tag_list=tag_list,
                               batchsize=args.batchsize,
                               constraints=constraints)
    else:
        doc = [l.strip() for l in fin]
        doc = [sentence for sentence in doc if len(sentence) > 0]
        tagged_doc = annotate_fun([[word for word in sent.split(' ')] for sent in doc],
                                  tokenize=args.tokenize)
        if args.tokenize:
            tagged_doc, doc = tagged_doc
        res = parser.parse_doc(doc,
                               probs=probs,
                               tag_list=tag_list,
                               batchsize=args.batchsize)

    if args.format == 'xml':
        output = etree.tostring(
            to_xml(res, tagged_doc), encoding='utf-8', pretty_print=True).decode('utf-8')
        print(output)
    elif args.format == 'jigg_xml':
        output = etree.tostring(
            to_jigg_xml(res, tagged_doc), encoding='utf-8', pretty_print=True).decode('utf-8')
        print(output)
    elif args.format == 'prolog':
        print(to_prolog(res, tagged_doc))
    elif args.format == 'html':
        print(to_mathml(res))
    elif args.format == 'jigg_xml_ccg2lambda':
        jigg_xml = to_jigg_xml(res, tagged_doc)
        templates = args.semantic_templates or SEMANTIC_TEMPLATES.get(args.lang)
        assert templates, f'--semantic-templates must be spcified for language: {args.lang}'
        result_xml_str, _ = ccg2lambda.parse(jigg_xml, str(templates), nbest=args.nbest)
        print(result_xml_str.decode('utf-8'))
    elif args.format == 'ccg2lambda':
        jigg_xml = to_jigg_xml(res, tagged_doc)
        templates = args.semantic_templates or SEMANTIC_TEMPLATES.get(args.lang)
        assert templates, f'--semantic-templates must be spcified for language: {args.lang}'
        _, formulas_list = ccg2lambda.parse(jigg_xml, str(templates), nbest=args.nbest)
        for i, (parsed, formulas) in enumerate(zip(res, formulas_list)):
            for j, ((tree, prob), formula) in enumerate(zip(parsed, formulas)):
                print(f'ID={i} log probability={prob:.4e}\n{formula}')
    elif args.format == 'conll':
        for i, parsed in enumerate(res):
            for tree, prob in parsed:
                print(f'# ID={i}\n# log probability={prob:.4e}\n{tree.conll()}')
    elif args.format == 'json':
        for i, (parsed, tokens) in enumerate(zip(res, tagged_doc), 1):
            for tree, prob in parsed:
                res = tree.json(tokens=tokens)
                res['id'] = i
                res['prob'] = prob
                json.dump(res, sys.stdout)
    elif args.format == 'auto':
        for i, (parsed, tokens) in enumerate(zip(res, tagged_doc), 1):
            for tree, prob in parsed:
                print(f'ID={i}, Prob={prob}')
                print(tree.auto(tokens=tokens))
    else:  # deriv, ja, ptb
        for i, parsed in enumerate(res, 1):
            for tree, prob in parsed:
                print(f'ID={i}, Prob={prob}')
                print(getattr(tree, args.format)())


def add_common_parser_arguments(parser):
    parser.add_argument('-m',
                        '--model',
                        help='path to model directory')
    parser.add_argument('-i',
                        '--input',
                        default=None,
                        help='a file with tokenized sentences in each line')
    parser.add_argument('-w',
                        '--weights',
                        default=None,
                        help='a file that contains weights (p_tag, p_dep)')
    parser.add_argument('--gpu',
                        type=int,
                        default=-1,
                        help='specify gpu id')
    parser.add_argument('--batchsize',
                        type=int,
                        default=32,
                        help='batchsize in supertagger')
    parser.add_argument('--nbest',
                        type=int,
                        default=1,
                        help='output N best parses')
    parser.add_argument('-I',
                        '--input-format',
                        default='raw',
                        choices=['raw', 'POSandNERtagged', 'json', 'partial'],
                        help='input format')
    parser.add_argument('--root-cats',
                        default=None,
                        help='allow only these categories at the root of a tree.')
    parser.add_argument('--unary-penalty',
                        default=0.1,
                        type=float,
                        help='penalty to use a unary rule')
    parser.add_argument('--beta',
                        default=0.00001,
                        type=float,
                        help='parameter used to filter categories with lower probabilities')
    parser.add_argument('--pruning-size',
                        default=50,
                        type=int,
                        help='use only the most probable supertags per word')
    parser.add_argument('--disable-beta',
                        action='store_true',
                        help='disable the use of the beta value')
    parser.add_argument('--disable-category-dictionary',
                        action='store_true',
                        help='disable a category dictionary that maps words to most likely supertags')
    parser.add_argument('--disable-seen-rules',
                        action='store_true',
                        help='')
    parser.add_argument('--max-length',
                        default=250,
                        type=int,
                        help='give up parsing a sentence that contains more words than this value')
    parser.add_argument('--max-steps',
                        default=10000000,
                        type=int,
                        help='give up parsing when the number of times of popping agenda items exceeds this value')
    parser.add_argument('--semantic-templates',
                        help='semantic templates used in "ccg2lambda" format output')
    parser.add_argument('--silent',
                        action='store_true')
    parser.set_defaults(func=main)

    subparsers = parser.add_subparsers()
    download_parser = subparsers.add_parser('download')
    download_parser.set_defaults(func=lambda args: download(args.lang))


if __name__ == '__main__':
    parser = argparse.ArgumentParser('A* CCG parser')
    parser.set_defaults(func=lambda _: parser.print_help())
    subparsers = parser.add_subparsers()

    english_parser = subparsers.add_parser('en')
    english_parser.set_defaults(lang='en')
    add_common_parser_arguments(english_parser)
    english_parser.add_argument('-a',
                                '--annotator',
                                default=None,
                                help='annotate POS, named entity, and lemmas using this library',
                                choices=english_annotator.keys())
    english_parser.add_argument('-f',
                                '--format',
                                default='auto',
                                choices=['auto', 'deriv', 'xml', 'conll', 'html', 'prolog', 'jigg_xml',
                                         'ptb', 'ccg2lambda', 'jigg_xml_ccg2lambda', 'json'],
                                help='output format')
    english_parser.add_argument('--tokenize',
                                action='store_true',
                                help='tokenize input sentences')
    english_parser.add_argument('--disfluency',
                                action='store_true',
                                help='perform disfluency detection')

    japanese_parser = subparsers.add_parser('ja')
    japanese_parser.set_defaults(lang='ja')
    add_common_parser_arguments(japanese_parser)
    japanese_parser.add_argument('-a',
                                 '--annotator',
                                 default='janome',
                                 help='annotate POS, named entity, and lemmas using this library',
                                 choices=japanese_annotator.keys())
    japanese_parser.add_argument('-f',
                                 '--format',
                                 default='ja',
                                 choices=['auto', 'deriv', 'ja', 'conll', 'html', 'jigg_xml',
                                          'ptb', 'ccg2lambda', 'jigg_xml_ccg2lambda', 'json'],
                                 help='output format')
    japanese_parser.add_argument('--pre-tokenized',
                                 dest='tokenize',
                                 action='store_false',
                                 help='the input is pre-tokenized (for running parsing experiments etc.)')
    japanese_parser.set_defaults(disfluency=False)

    args = parser.parse_args()
    args.func(args)






#! -*- coding: utf-8 -*-

from __future__ import print_function
from setuptools import Extension, setup, find_packages
import tempfile
import subprocess
import shutil
import sys
import subprocess
import os
import distutils
import platform
import contextlib


try:
    from Cython.Build import build_ext
except ImportError:
    sys.exit("""Could not import Cython, which is required to build depccg extension modules.
Please install cython and numpy prior to installing depccg.""")

try:
    import numpy
except ImportError:
    sys.exit("""Could not import numpy, which is required to build the extension modules.
Please install cython and numpy prior to installing depccg.""")

here = os.path.abspath(os.path.dirname(__file__))

install_requires = [line.strip() for line in open(
                    os.path.join(here, 'requirements.txt'))]

with open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

@contextlib.contextmanager
def chdir(new_dir):
    old_dir = os.getcwd()
    try:
        os.chdir(new_dir)
        sys.path.insert(0, new_dir)
        yield
    finally:
        del sys.path[0]
        os.chdir(old_dir)


COMPILE_OPTIONS = ['-O3', '-Wno-strict-prototypes', '-Wno-unused-function', '-std=c++11']

LINK_OPTIONS = []

CPP_OPTIONS = []

USE_OPENMP_DEFAULT = '0' if sys.platform != 'darwin' else None
if os.environ.get('USE_OPENMP', USE_OPENMP_DEFAULT) == '1':
    COMPILE_OPTIONS.append('-fopenmp')
    LINK_OPTIONS.append('-fopenmp')
    CPP_OPTIONS.append('-fopenmp')

if sys.platform == 'darwin':
    COMPILE_OPTIONS.append('-stdlib=libc++')
    LINK_OPTIONS.append('-lc++')
    # g++ (used by unix compiler on mac) links to libstdc++ as a default lib.
    # See: https://stackoverflow.com/questions/1653047/avoid-linking-to-libstdc
    LINK_OPTIONS.append('-nodefaultlibs')


cpp_sources = ['depccg.cpp',
               'cat.cpp',
               'cacheable.cpp',
               'chart.cpp',
               'combinator.cpp',
               'feat.cpp',
               'tree.cpp',
               'utils.cpp']

c_sources = ['morpha.c']

pyx_modules = ['depccg.parser',
               'depccg.tree',
               'depccg.cat',
               'depccg.combinator',
               'depccg.utils',
               'depccg.morpha']


def clean():
    depccg_dir = os.path.join(here, 'depccg')
    for name in os.listdir(depccg_dir):
        file_path = os.path.join(depccg_dir, name)
        if any(file_path.endswith(ext) for ext in ['.so', '.cpp', '.c']):
            os.unlink(file_path)
    with chdir('cpp'):
        subprocess.call(["make", "clean"], env=os.environ)


def generate_cython(here, source):
    print('Cythonizing sources')
    p = subprocess.call([sys.executable,
                         os.path.join(here, 'bin', 'cythonize.py'),
                         source], env=os.environ)
    if p != 0:
        raise RuntimeError('Running cythonize failed')


def generate_cpp(options):
    options = ' '.join(options)
    options = f'OPTIONS={options}'
    with chdir('cpp'):
        p = subprocess.call(["make", options], env=os.environ)
        if p != 0:
            raise RuntimeError('Running cythonize failed')


ext_modules = [
        Extension(pyx,
                  [pyx.replace('.', '/') + '.cpp'],
                  include_dirs=['.', numpy.get_include(), 'cpp'],
                  extra_compile_args=COMPILE_OPTIONS,
                  extra_link_args=LINK_OPTIONS +
                  [os.path.join('cpp', cpp.replace('cpp', 'o')) for cpp in cpp_sources] +
                  [os.path.join('cpp', cpp.replace('c', 'o')) for cpp in c_sources],
                  language='c++')
        for pyx in pyx_modules]


if len(sys.argv) > 1 and sys.argv[1] == 'clean':
    clean()
else:
    generate_cpp(CPP_OPTIONS)

    generate_cython(here, 'depccg')

    setup(
        name="depccg",
        version="1.0.2",  # NOQA
        description='A parser for natural language based on combinatory categorial grammar',
        long_description=long_description,
        author='Masashi Yoshikawa',
        author_email='yoshikawa.masashi.yh8@is.naist.jp',
        url='https://github.com/masashi-y/depccg',
        license='MIT License',
        packages=find_packages(),
        package_data={'depccg': ['models/*']},
        ext_modules=ext_modules,
        cmdclass={"build_ext": build_ext},
        scripts=['bin/depccg_en', 'bin/depccg_ja'],
        install_requires=install_requires,
        classifiers=[
            'License :: OSI Approved :: MIT License',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
        ],
        zip_safe=False,
    )
